#include <coemain.h>
#include "caffeineappview.h"
#include <flogger.h>
#include <hal.h>
#include <eikenv.h>
#include <aknutils.h>
#include <aknnotewrappers.h>
#include <eikapp.h>
#include "symbiancaffeine.h"
#include "caffeineappui.h"

void debug(const char *str) {
	TBuf<200> buf;
	while (*str)
		buf.Append(*str++);
	RFileLogger::WriteFormat(_L("debug"), _L("caffeine.txt"), EFileLoggingModeAppend, buf);
}

CCaffeineAppView* CCaffeineAppView::NewL(const TRect& aRect) {
	CCaffeineAppView* self = CCaffeineAppView::NewLC(aRect);
	CleanupStack::Pop(self);
	return self;
}

CCaffeineAppView* CCaffeineAppView::NewLC(const TRect& aRect) {
	CCaffeineAppView* self = new(ELeave) CCaffeineAppView();
	CleanupStack::PushL(self);
	self->ConstructL(aRect);
	return self;
}

CCaffeineAppView::CCaffeineAppView() {
	iInCallback = EFalse;
}

CCaffeineAppView::~CCaffeineAppView() {
	if (iCaffeine)
		iCaffeine->stopAudio();
	delete iCaffeine;

	delete iBitmap;
	delete iDevice;
	delete iGc;
	delete iDrawer;
	delete iPeriodic;

	delete [] iFrameTick;
}

void CCaffeineAppView::ConstructL(const TRect& aRect) {
	iDSAStarted = EFalse;
	iDrawMode = EDSA;

	iFrameTickLength = 50;
	iFrameTick = new(ELeave) TInt[iFrameTickLength];
	for (int i = 0; i < iFrameTickLength; i++)
		iFrameTick[i] = 0;

	CreateWindowL();
	SetExtentToWholeScreen();
	ActivateL();

	int bpp = 0;
	switch (iEikonEnv->ScreenDevice()->DisplayMode()) {
	case EColor4K:
		bpp = 12;
		break;
	case EColor64K:
		bpp = 16;
		break;
	case EColor16MU:
		bpp = 32;
		break;
	default:
		break;
	}
	if (bpp > 0)
		iCaffeine = new(ELeave) SymbianCaffeine(Rect().Width(), Rect().Height(), bpp, iBitmap->ScanLineLength(Rect().Width(), iBitmap->DisplayMode()));

	if (HAL::Get(HALData::ESystemTickPeriod, iPeriod) != KErrNone)
		iPeriod = 1000/64;
	else
		iPeriod /= 1000;


	iPeriodic = CPeriodic::NewL(CActive::EPriorityIdle);
	iStarted = EFalse;
	SetForeground(ETrue);

	iFrame = 0;
	iLastFPSFrame = 0;
	iLastFPSTime = (TUint)-1;
	iTime = 0;
	iLastFPS = 0;
	iLastResetTick = 0;

}

void CCaffeineAppView::HandleResourceChange(TInt aType) {
	CCoeControl::HandleResourceChange(aType);
	if (aType == KEikDynamicLayoutVariantSwitch) {
//		SetExtentToWholeScreen();
//		if (iCaffeine)
//			iCaffeine->EnableMusic(EFalse);

		CAknInformationNote* informationNote = new(ELeave) CAknInformationNote(ETrue);
		_LIT(message, "Layout switch not supported");
		iInCallback = ETrue;
		informationNote->ExecuteLD(message);
		static_cast<CCaffeineAppUi*>(iEikonEnv->EikAppUi())->DoExit();
	}
}

void CCaffeineAppView::SizeChanged() {
	SetupBitmap();
}

void CCaffeineAppView::SetupBitmap() {
	delete iBitmap;
	delete iDevice;
	delete iGc;

	iBitmap = new(ELeave) CWsBitmap(iEikonEnv->WsSession());
	iBitmap->Create(Rect().Size(), iEikonEnv->ScreenDevice()->DisplayMode());

	iDevice = CFbsBitmapDevice::NewL(iBitmap);
	TInt err = iDevice->CreateContext(iGc);
	if (err != KErrNone) {
		_LIT(KPanicMsg, "CFbsBitmapDevice::CreateContext");
		User::Panic(KPanicMsg, err);
	}
	iGc->SetBrushColor(KRgbBlack);
	iGc->Clear(TRect(TPoint(0, 0), iBitmap->SizeInPixels()));

	SetDrawMode(iDrawMode, ETrue);
}

void CCaffeineAppView::SetDrawMode(TDrawMode aMode, TBool aReset) {
	if (aMode == iDrawMode && !aReset)
		return;

	delete iDrawer;
	iDrawer = NULL;

	iDrawMode = aMode;

	TInt err;
	switch (iDrawMode) {
	case EDoubleBuf:
		break;

	case EDSA:
		iDrawer = CDirectScreenAccess::NewL(iEikonEnv->WsSession(), *(iEikonEnv->ScreenDevice()), Window(), *this);
		iEikonEnv->WsSession().Flush();
		TRAP(err, iDrawer->StartL());
		if (err == KErrNone) {
			iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
			iDSAStarted = ETrue;
		}
		break;
	}


}


void CCaffeineAppView::SetForeground(TBool aForeground) {
	iForeground = aForeground;
	SetDrawing(aForeground);
}

void CCaffeineAppView::SetDrawing(TBool aDrawing) {
	if (!iForeground)
		aDrawing = EFalse;
	if (aDrawing) {
		if (!iStarted) {
			iPeriodic->Start(100, 100, TCallBack(CCaffeineAppView::StaticCallback, this));
			iLastTick = User::TickCount();
		}
	} else {
		if (iStarted)
			iPeriodic->Cancel();
	}
	iStarted = aDrawing;
}

void CCaffeineAppView::Draw(const TRect& aRect) const {
	CWindowGc& gc = SystemGc();
	gc.BitBlt(TPoint(0, 0), iBitmap);
}

void CCaffeineAppView::Update() {
	if (!iForeground) return;
	TInt err = KErrNone;
	if (!iDSAStarted) {
		TRAP(err, iDrawer->StartL());
		if (err == KErrNone) {
			iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
			iDSAStarted = ETrue;
		}
	}
	if (err == KErrNone) {
		iDrawer->Gc()->BitBlt(TPoint(0, 0), iBitmap);
	}
}

void CCaffeineAppView::Restart(RDirectScreenAccess::TTerminationReasons aReason) {
	SetDrawing(ETrue);
}

void CCaffeineAppView::AbortNow(RDirectScreenAccess::TTerminationReasons aReason) {
	SetDrawing(EFalse);
	iDSAStarted = EFalse;
}


TInt CCaffeineAppView::StaticCallback(TAny* aArg) {
	CCaffeineAppView* view = (CCaffeineAppView*) aArg;
	if (view->iInCallback) return 0;
	view->iInCallback = ETrue;
	view->Callback();
	view->iInCallback = EFalse;
	return 0;
}

void CCaffeineAppView::Callback() {
	iFrame++;

	TInt tick = User::TickCount();
	iRelTick += tick - iLastTick;
	iTime = iRelTick*iPeriod;
	iLastTick = tick;


	if (iCaffeine == NULL) {
		CAknInformationNote* informationNote = new(ELeave) CAknInformationNote(ETrue);
		_LIT(message, "Unsupported display mode");
		informationNote->ExecuteLD(message);
		static_cast<CCaffeineAppUi*>(iEikonEnv->EikAppUi())->DoExit();
		return;
	}

	iCaffeine->update(iTime);
#ifdef EKA2
	iBitmap->LockHeap();
#endif
	iCaffeine->draw((unsigned char*) iBitmap->DataAddress());
#ifdef EKA2
	iBitmap->UnlockHeap();
#endif
	// give other threads a chance also
	// makes the sound start faster
	User::After(1*1000);




	iFrameTick[iFrame % iFrameTickLength] = iRelTick;

	if (iDrawMode == EDoubleBuf) {
		DrawNow();
	} else if (iDrawMode == EDSA) {
		Update();
	}


	if (iLastFPSTime == (TUint)-1)
		iLastFPSTime = iTime;

	if ((tick - iLastResetTick)*iPeriod >= 3000) {
		User::ResetInactivityTime();
		iLastResetTick = tick;
	}

	if (iTime - iLastFPSTime > 1000) {
		iLastFPS = 1000*(iFrame - iLastFPSFrame)/(iTime - iLastFPSTime);
		RFileLogger::WriteFormat(_L("debug"), _L("caffeine.txt"), EFileLoggingModeAppend, _L("%d fps"), iLastFPS);
		iLastFPSTime = iTime;
		iLastFPSFrame = iFrame;
	}


	if (iCaffeine->isFinished())
		static_cast<CCaffeineAppUi*>(iEikonEnv->EikAppUi())->DoExit();

}

TKeyResponse CCaffeineAppView::OfferKeyEventL(const TKeyEvent& aKeyEvent, TEventCode aType) {
	return EKeyWasNotConsumed;
}

