#include <stdlib.h>
#include <math.h>
#include "rotoplasma.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

RotoPlasma::RotoPlasma(int w, int h, int b, int p) {
	width = w;
	height = h;
	bpp = b;
	pitch = p;
	texture1 = texture2 = NULL;
	palette = NULL;
	step = 0;
}

RotoPlasma::~RotoPlasma(void) {
	free(sintable);
	free(texture1);
	free(texture2);
	free(palette);
}

float RotoPlasma::fastsin(float f) {
	int i = int(512*f/(2*M_PI));
	i &= 0x01FF;
	return sintable[i];
}

float RotoPlasma::fastcos(float f) {
	int i = int(512*f/(2*M_PI));
	i += 128;
	i &= 0x01FF;
	return sintable[i];
}

void RotoPlasma::init(void) {
	if (step == 0) {
		angleA = atan2(-height,-width);
		angleB = atan2(-height,width);
		angleC = atan2(height,-width);

		texture1 = (unsigned char*) malloc(sizeof(char)*512*512);
		texture2 = (unsigned char*) malloc(sizeof(char)*512*512);


		sintable = (float*) malloc(sizeof(float)*512);
		for (int i = 0; i < 512; i++)
			sintable[i] = sin(2*M_PI*i/512);

		palette = createPalette();

		step++;
		return;
	}

	int minx = 128 * ((step - 1)%4);
	int miny = 128 * ((step - 1)/4);
	int maxx = minx + 128;
	int maxy = miny + 128;
	for (int y = miny; y < maxy; y++) {
		unsigned char *ptr1 = texture1 + y*512 + minx;
		unsigned char *ptr2 = texture2 + y*512 + minx;
		for (int x = minx; x < maxx; x++) {
			*ptr1++ = (int) (64+63*fastsin(10+x/(42+15*fastsin((float)y/27)))*fastcos(15+y/(38+10*fastcos((float)x/25))));
			*ptr2++ = (int) (64+63*fastsin(x/(37+13*fastcos((float)y/23)))*fastcos(y/(40+8*fastsin((float)x/28))));
		}
	}

	if (step == 16) {
		free(sintable);
		sintable = NULL;
	}
	step++;

}

bool RotoPlasma::isInited(void) {
	return step >= 17;
}

unsigned int *RotoPlasma::createPalette(void) {
	unsigned int *pal = (unsigned int*) malloc(sizeof(int)*256);
	unsigned int *ptr = pal;
	for (int i = 0; i < 256; i++)
		*ptr++ = (255-i)<<16 | i<<8;

	if (bpp == 12) { 
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>12) & 0x0F00) | ((pal[i]>>8) & 0x00F0) | ((pal[i]>>4) & 0x000F);
	} else if (bpp == 16) {
		for (int i = 0; i < 256; i++)
			pal[i] = ((pal[i]>>8) & 0xF800) | ((pal[i]>>5) & 0x07E0) | ((pal[i]>>3) & 0x001F); 
	}

	return pal;
}

void RotoPlasma::draw(unsigned char *buffer, int time, bool bilinear) {
	float radius = 75+45*sin((float)time/2000);
	float angle;
	if (time<15000)
		angle = M_PI*time/3000;
	else
		angle = M_PI*(30000-time)/3000;

	float centerX = 256+10*cos((float)time/1000);
	float centerY = 256+10*sin(10+(float)time/800);

	float radiusA = (90+10*sin(2+(float)time/1000))/100;
	float radiusB = (80+20*sin(3+(float)time/1200))/100;
	float radiusC = (85+15*sin(4+(float)time/850))/100;

	int ax = (int) ((1<<16)*(centerX+radius*radiusA*cos(angleA+angle)));
	int ay = (int) ((1<<16)*(centerY+radius*radiusA*sin(angleA+angle)));
	int bx = (int) ((1<<16)*(centerX+radius*radiusB*cos(angleB+angle)));
	int by = (int) ((1<<16)*(centerY+radius*radiusB*sin(angleB+angle)));
	int cx = (int) ((1<<16)*(centerX+radius*radiusC*cos(angleC+angle)));
	int cy = (int) ((1<<16)*(centerY+radius*radiusC*sin(angleC+angle)));

	int dxdx = (bx-ax)/width;
	int dydx = (by-ay)/width;
	int dxdy = (cx-ax)/height;
	int dydy = (cy-ay)/height;

	int offsetX1 = (int) ((1<<16)*120*sin((float)(3000+time)/2000));
	int offsetY1 = (int) ((1<<16)*120*cos((float)(4000+time)/3000));
	int offsetX2 = (int) ((1<<16)*120*cos((float)(11000+time)/2500));
	int offsetY2 = (int) ((1<<16)*120*sin((float)(13000+time)/2800));

	if (bilinear) {
		int px11,px12,px13,px14,px21,px22,px23,px24;
		int fx1,fx2,fy1,fy2;
		if (bpp <= 16) {
			unsigned short *outptr = (unsigned short*) buffer;
			int spitch = pitch/2;
			for (int y = 0; y < height; y++) {
				cx = ax;
				cy = ay;
				for (int x = 0; x < width; x++) {
					px11 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+((cx+offsetX1)>>16)];
					px12 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+((cx+offsetX1)>>16)+1];
					px13 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+0x01FF+((cx+offsetX1)>>16)];
					px14 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+0x01FF+((cx+offsetX1)>>16)+1];
					px21 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+((cx+offsetX2)>>16)];
					px22 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+((cx+offsetX2)>>16)+1];
					px23 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+0x01FF+((cx+offsetX2)>>16)];
					px24 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+0x01FF+((cx+offsetX2)>>16)+1];
					fx1 = ((cx+offsetX1)&0xFFFF)>>8;
					fy1 = ((cy+offsetY1)&0xFFFF)>>8;
					fx2 = ((cx+offsetX2)&0xFFFF)>>8;
					fy2 = ((cy+offsetY2)&0xFFFF)>>8;
					*outptr++ = palette[(
						px11*((0xFF-fx1)*(0xFF-fy1)) +
						px12*((fx1)*(0xFF-fy1)) +
						px13*((0xFF-fx1)*(fy1)) +
						px14*((fx1)*(fy1)) +
						px21*((0xFF-fx2)*(0xFF-fy2)) +
						px22*((fx2)*(0xFF-fy2)) +
						px23*((0xFF-fx2)*(fy2)) +
						px24*((fx2)*(fy2))
					)>>16];
					cx += dxdx;
					cy += dydx;
				}
				ax += dxdy;
				ay += dydy;
				outptr += spitch - width;
			}
		} else {
			unsigned int *outptr = (unsigned int*) buffer;
			int ipitch = pitch/4;
			for (int y = 0; y < height; y++) {
				cx = ax;
				cy = ay;
				for (int x = 0; x < width; x++) {
					px11 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+((cx+offsetX1)>>16)];
					px12 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+((cx+offsetX1)>>16)+1];
					px13 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+0x01FF+((cx+offsetX1)>>16)];
					px14 = texture1[(((cy+offsetY1)>>7)&0x03FE00)+0x01FF+((cx+offsetX1)>>16)+1];
					px21 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+((cx+offsetX2)>>16)];
					px22 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+((cx+offsetX2)>>16)+1];
					px23 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+0x01FF+((cx+offsetX2)>>16)];
					px24 = texture2[(((cy+offsetY2)>>7)&0x03FE00)+0x01FF+((cx+offsetX2)>>16)+1];
					fx1 = ((cx+offsetX1)&0xFFFF)>>8;
					fy1 = ((cy+offsetY1)&0xFFFF)>>8;
					fx2 = ((cx+offsetX2)&0xFFFF)>>8;
					fy2 = ((cy+offsetY2)&0xFFFF)>>8;
					*outptr++ = palette[(
						px11*((0xFF-fx1)*(0xFF-fy1)) +
						px12*((fx1)*(0xFF-fy1)) +
						px13*((0xFF-fx1)*(fy1)) +
						px14*((fx1)*(fy1)) +
						px21*((0xFF-fx2)*(0xFF-fy2)) +
						px22*((fx2)*(0xFF-fy2)) +
						px23*((0xFF-fx2)*(fy2)) +
						px24*((fx2)*(fy2))
					)>>16];
					cx += dxdx;
					cy += dydx;
				}
				ax += dxdy;
				ay += dydy;
				outptr += ipitch - width;
			}
		}
	} else {
		if (bpp <= 16) {
			unsigned short *outptr = (unsigned short*) buffer;
			int spitch = pitch/2;
			for (int y = 0; y < height; y++) {
				cx = ax;
				cy = ay;
				for (int x = 0; x < width; x++) {
					*outptr++ = palette[
						texture1[(((cy+offsetY1)>>7)&0x03FE00)+((cx+offsetX1)>>16)] +
						texture2[(((cy+offsetY2)>>7)&0x03FE00)+((cx+offsetX2)>>16)]
					];
					cx += dxdx;
					cy += dydx;
				}
				ax += dxdy;
				ay += dydy;
				outptr += spitch - width;
			}
		} else {
			unsigned int *outptr = (unsigned int*) buffer;
			int ipitch = pitch/4;
			for (int y = 0; y < height; y++) {
				cx = ax;
				cy = ay;
				for (int x = 0; x < width; x++) {
					*outptr++ = palette[
						texture1[(((cy+offsetY1)>>7)&0x03FE00)+((cx+offsetX1)>>16)] +
						texture2[(((cy+offsetY2)>>7)&0x03FE00)+((cx+offsetX2)>>16)]
					];
					cx += dxdx;
					cy += dydx;
				}
				ax += dxdy;
				ay += dydy;
				outptr += ipitch - width;
			}
		}
	}
}

