#include "wavestream.h"
#include <aknnotewrappers.h>
#include <e32math.h>
#include <flogger.h>
#include <stdlib.h>


CWaveStream* CWaveStream::NewL(const TDesC& aName, MWaveGenerator* aGen, TInt aVolume) {
	CWaveStream* self = new(ELeave) CWaveStream();
	CleanupStack::PushL(self);
	self->ConstructL(aName, aGen, aVolume);
	CleanupStack::Pop(self);
	return self;
}
	
CWaveStream::CWaveStream() {
}

void CWaveStream::ConstructL(const TDesC& aName, MWaveGenerator* aGen, TInt aVolume) {
	if (aVolume < 0)
		aVolume = 0;
	else if (aVolume > 100)
		aVolume = 100;
	iShare = new(ELeave) CStreamShareData();
	iShare->iGen = aGen;
	iShare->iMutex.CreateLocal();
	iShare->iAliveMutex.CreateLocal();
	iShare->iCommandSemaphore.CreateLocal(0);
	iShare->iName = aName;
	iShare->iStopped = EFalse;
	iShare->iVolume = aVolume;

	User::LeaveIfError(iThread.Create(aName, CWaveThread::ThreadFunc, KDefaultStackSize, NULL, iShare));
	iThread.SetPriority(EPriorityRealTime);

	iThread.Resume();
}

void CWaveStream::SetGenerator(MWaveGenerator* aGen) {
	iShare->iMutex.Wait();
	iShare->iGen = aGen;
	iShare->iMutex.Signal();
}

void CWaveStream::SetUserData(TAny* aData) {
	iShare->iMutex.Wait();
	iShare->iUserData = aData;
	iShare->iMutex.Signal();
}
	
void CWaveStream::LockShared() {
	iShare->iMutex.Wait();
}

void CWaveStream::UnlockShared() {
	iShare->iMutex.Signal();
}

CWaveStream::~CWaveStream() {

	if (iShare) {
		SendCommand(EStreamKill);
	
		iShare->iAliveMutex.Wait();
		iShare->iAliveMutex.Signal();
		
		iShare->iMutex.Close();
		iShare->iAliveMutex.Close();
		iShare->iCommandSemaphore.Close();
		delete iShare;
	}
	
	iThread.Close();
}

void CWaveStream::Start() {
	SendCommand(EStreamStart);
}

void CWaveStream::Stop() {
	SendCommand(EStreamStop);
}

CStreamShareData::CStreamShareData() {
}

void CWaveStream::SendCommand(TStreamCommand aCommand) {
	iShare->iCommandSemaphore.Wait();
	if (iShare->iStopped)
		User::Panic(iShare->iName, 6);

	iShare->iMutex.Wait();
	iShare->iCommand = aCommand;
	iShare->iMutex.Signal();
	iThread.RequestComplete(iShare->iStatusPtr, KErrNone);
}

void CWaveStream::SetVolume(TInt aVolume) {
	if (aVolume < 0)
		aVolume = 0;
	else if (aVolume > 100)
		aVolume = 100;

	iShare->iMutex.Wait();
	iShare->iVolume = aVolume;
	iShare->iMutex.Signal();

	SendCommand(EStreamSetVolume);
}







TInt CWaveThread::ThreadFunc(TAny *aPtr) {
	CStreamShareData* share = static_cast<CStreamShareData*>(aPtr);
	share->iAliveMutex.Wait();
	CTrapCleanup* cleanup = CTrapCleanup::New();
	TRAPD(err, ThreadFuncL(aPtr));
	delete cleanup;
	share->iStopped = ETrue;
	share->iCommandSemaphore.Signal();
	share->iAliveMutex.Signal();
	CloseSTDLIB();
	return err;
}
	
void CWaveThread::ThreadFuncL(TAny *aPtr) {
	CActiveScheduler* scheduler = new(ELeave) CActiveScheduler;
	CleanupStack::PushL(scheduler);
	CActiveScheduler::Install(scheduler);
	
	CWaveThread* waveThread = CWaveThread::NewL(static_cast<CStreamShareData*>(aPtr));
	CleanupStack::PushL(waveThread);

	CActiveScheduler::Start();
	CleanupStack::PopAndDestroy(waveThread);
	CleanupStack::PopAndDestroy(scheduler);
}


void CWaveThread::HandleCommand() {
	switch (iShare->iCommand) {
	case EStreamStart:
		Start();
		break;
	case EStreamStop:
		Stop();
		break;
	case EStreamSetVolume:
		UpdateVolume();
		break;
	case EStreamKill:
	default:
		iStopping = ETrue;
		CActiveScheduler::Stop();
		break;
	}
}


CWaveThread::CWaveThread(CStreamShareData* aShare) : CActive(EPriorityStandard) {
	iShare = aShare;
	iStopping = EFalse;
}


CWaveThread::~CWaveThread() {
	Cancel();
	if (iOutput) {
		iOutput->Stop();
		delete iOutput;
		iOutput = NULL;
	}
	delete iBuf1;
}


CWaveThread* CWaveThread::NewL(CStreamShareData* aShare) {
	CWaveThread* self = new(ELeave) CWaveThread(aShare);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop(self);
	return self;
}

void CWaveThread::ConstructL() {
	CActiveScheduler::Add(this);
	StartRequest();

	iBuf1 = HBufC8::NewMaxL(800);
}

void CWaveThread::StartRequest() {
	iStatus = KRequestPending;
	SetActive();
	iShare->iStatusPtr = &iStatus;
	iShare->iCommandSemaphore.Signal();
}

void CWaveThread::DoCancel() {
	User::RequestComplete(iShare->iStatusPtr, KErrNone);
}

void CWaveThread::RunL() {
	if (iStatus == KErrNone) {
		HandleCommand();
		if (!iStopping)
			StartRequest();
	}
}

void CWaveThread::Start() {
	if (iOutput) return;
	iSettings.iSampleRate = TMdaAudioDataSettings::ESampleRate16000Hz;
	iSettings.iChannels = TMdaAudioDataSettings::EChannelsMono;
	iShare->iGen->SetSampleRate(16000);
	TRAPD(err, iOutput = CMdaAudioOutputStream::NewL(*this));
	if (err != KErrNone)
		User::Panic(_L("CMAOS::NewL"), err);
	if (iOutput)
		iOutput->Open(&iSettings);
}

void CWaveThread::Stop() {
	if (!iOutput) return;
	iOutput->Stop();
	delete iOutput;
	iOutput = NULL;
}

void CWaveThread::UpdateVolume() {
	if (!iOutput)
		return;

	iOutput->SetVolume(iOutput->MaxVolume()*iShare->iVolume/100);
}

void CWaveThread::RenderData(TPtr8 aPtr) {
	TInt16* ptr = (TInt16*) aPtr.Ptr();
	TInt len = aPtr.Length()/2;
	iShare->iMutex.Wait();
	iShare->iGen->RenderData(iShare->iUserData, ptr, len);
	iShare->iMutex.Signal();
}

void CWaveThread::MaoscOpenComplete(TInt aError) {
	if (aError == KErrNone) {
		iOutput->SetVolume(iOutput->MaxVolume()*iShare->iVolume/100);
		iOutput->SetPriority(EMdaPriorityMax, EMdaPriorityPreferenceTime);

		RenderData(iBuf1->Des());
		TRAPD(err, iOutput->WriteL(*iBuf1));
		if (err != KErrNone)
			User::Panic(_L("CMAOS::WriteL"), err);

	} else {
		User::Panic(_L("CMAOS::MaoscOpenComplete"), aError);
	}
}

void CWaveThread::MaoscBufferCopied(TInt aError, const TDesC8& aBuffer) {
	if (aError == KErrAbort)
		return;
	if (aError != KErrNone)
		User::Panic(_L("CMAOS::MaoscBufferCopied"), aError);

	RenderData(((HBufC8*)&aBuffer)->Des());
	TRAPD(err, iOutput->WriteL(aBuffer));
	if (err != KErrNone)
		User::Panic(_L("CMAOS::WriteL"), err);
}

void CWaveThread::MaoscPlayComplete(TInt aError) {
	if (aError == KErrNone || aError == KErrUnderflow || aError == KErrNotReady) {
		RenderData(iBuf1->Des());
		iOutput->SetVolume(iOutput->MaxVolume()/4);
		TRAPD(err, iOutput->WriteL(*iBuf1));
		if (err != KErrNone)
			User::Panic(_L("CMAOS::WriteL"), err);
	} else if (aError == KErrCancel) {
	} else {
		User::Panic(_L("CMAOS::MaoscPlayComplete"), aError);
	}
}



